"""The `Client` class is your interface to the quantum computing platform."""
import traceback

from typeguard import typechecked

from .client import QciConnectClient
from .exceptions import QciConnectClientError, ResultError
from .methods import Methods
from .qpus import QpuByAlias
from .result_handling import Results


class Client:
    """When you create a `Client` object, it connects to a QCI Connect service.

    To connect to the official QCI Connect platform, provide your authentication token (found in the [webfrontend](https://connect.qci.dlr.de)):

        from qciconnect import Client
        client = Client(token="...")

    To use your own QCI Connect service, specify the server address:

        client = Client(server="https://my-own-server.com/")

    The `Client` object has two attributes:
    - `methods`: Compiler passes and postprocessing tools provided by the platform.
    - `qpus`: contains all the quantum processing units (including simulators) available on the platform.
    """ # noqa: E501

    _DEFAULT_SERVER = "https://connect.qci.dlr.de/"
    _LOCALHOST = "http://127.0.0.1:8000/"

    @typechecked
    def __init__(
        self,
        /,
        token: str | None = None,
        *,
        server: str | None = None
    ):
        """Constructs a Mageia/Client object."""
        if token is None and server is None:
            print(f"Warning! No server and no credentials given.\n"
                  f"In order to use the QCI Connect server {self._DEFAULT_SERVER} "
                  f"you need to provide username and password.\n"
                  f"Falling back to {self._LOCALHOST} as server.")
            server = self._LOCALHOST
        if token is not None and server is None:
            server = self._DEFAULT_SERVER

        try:
            self._client = QciConnectClient(
                server_address=server, 
                token=token)
            if server == self._DEFAULT_SERVER:
                self._update_tac_and_privacy_status()
                if not self.tacs_accepted:
                    print(
                        f"Terms and conditions not accepted. Please accept them on "
                        f"{self._DEFAULT_SERVER}."
                    )
                if not self.privacy_notice_accepted:
                    print(
                        f"Privacy notice not accepted. Please accept the privacy notice on "
                        f"{self._DEFAULT_SERVER} using a web browser."
                    )

            self._update_methods()
            self._update_qpus()
            self._update_results()
        except (QciConnectClientError, ResultError) as e:
            print(e)
        except Exception as e:  # noqa: BLE001
            if (
                server.__contains__("localhost")
                or server.__contains__("0.0.0.0")
                or server.__contains__("127.0.0.1")
            ):
                print(
                    "Please run `docker compose up` before connecting to your backend. \n"
                    "For further information see also "
                    "https://gitlab.dlr.de/qci-connect/partners/all/qciconnect-sdk"
                )
            else:
                print(f"Connection to {server} could not be established.")
                traceback.print_exception(e)
                print(e, type(e))


    @property
    def server_address(self) -> str:
        """Returns URL of the platform."""
        return self._client.server_address

    @server_address.setter
    def server_address(self, value: str):
        """Sets the URL of the platform and retrieves available methods and QPUs.

        :param value: URL of the platform that shall be used.
        """
        self._client.server_address = value
        self._update_methods()
        self._update_qpus()
        self._update_results()

    def _update_methods(self):
        """Updates the methods/compiler passes available on the platform."""
        compiler_list = self._client.get_available_compilers()
        self.methods = Methods(compiler_list, self._client)

    def _update_qpus(self):
        """Updates the quantum processing units available on the platform."""
        self.qpus = QpuByAlias(self._client)

    def _update_results(self):
        """Updates the results of the jobs available on the platform."""
        self.results = Results(self._client)
        self.results.refresh()

    def _update_tac_and_privacy_status(self):
        """Updates the terms and conditions and privacy notice status of the platform."""
        self.tacs_accepted = self._client.get_tac_status()
        self.privacy_notice_accepted = self._client.get_privacy_notice_status()



"""Unit tests for conversion between NumPy arrays and JSON."""

import pytest
from numpy import array
from qciconnect_common.qciconnect_conversion.json_conversion import JSONNumpyConverter

conv = JSONNumpyConverter()


class TestJSONNumpyConverter:
    """Contains unit tests for conversion between NumPy arrays and JSON."""

    @pytest.mark.parametrize(
        ("numpy_array_input", "json_list_input"),
        [
            (
                array([0, 1, 2]),
                "[0, 1, 2]",
            ),
            (
                array([0.1, 1.2, 2.3]),
                "[0.1, 1.2, 2.3]",
            ),
            (
                array([]),
                "[]",
            ),
            (
                array([[0.1, 1.2, 2.3], [0.1, 1.2, 2.3]]),
                "[[0.1, 1.2, 2.3], [0.1, 1.2, 2.3]]",
            ),
            (
                array(
                    [
                        [0, 1, 1, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 1, 0, 0, 1],
                        [1, 1, 1, 0, 1],
                        [0, 0, 0, 0, 0],
                        [1, 1, 0, 0, 1],
                    ]
                ),
                "[[0, 1, 1, 0, 0], [0, 0, 0, 0, 0], [0, 1, 0, 0, 1], "
                "[1, 1, 1, 0, 1], [0, 0, 0, 0, 0], [1, 1, 0, 0, 1]]",
            ),
        ],
    )
    def test_nparray_to_json_list(self, numpy_array_input, json_list_input):
        """Test conversion from NumPy array to JSON compatible list.

        Args:
            numpy_array_input (nparray): Test input array.
            json_list_input (str): Expected JSON list output.
        """
        converted_input = conv.nparray_to_json_list(numpy_array_input)
        assert converted_input == json_list_input

    @pytest.mark.parametrize(
        ("numpy_array_input", "json_list_input"),
        [
            (
                array([0, 1, 2]),
                "[0, 1, 2]",
            ),
            (
                array([0.1, 1.2, 2.3]),
                "[0.1, 1.2, 2.3]",
            ),
            (
                array([]),
                "[]",
            ),
            (
                array([[0.1, 1.2, 2.3], [0.1, 1.2, 2.3]]),
                "[[0.1, 1.2, 2.3], [0.1, 1.2, 2.3]]",
            ),
            (
                array(
                    [
                        [0, 1, 1, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 1, 0, 0, 1],
                        [1, 1, 1, 0, 1],
                        [0, 0, 0, 0, 0],
                        [1, 1, 0, 0, 1],
                    ]
                ),
                "[[0, 1, 1, 0, 0], [0, 0, 0, 0, 0], [0, 1, 0, 0, 1], "
                "[1, 1, 1, 0, 1], [0, 0, 0, 0, 0], [1, 1, 0, 0, 1]]",
            ),
        ],
    )
    def test_json_list_to_nparray(self, numpy_array_input, json_list_input):
        """Test conversion from JSON compatible list to NumPy array.

        Args:
            numpy_array_input (nparray): Expected output array.
            json_list_input (str): JSON list input.
        """
        converted_input = conv.json_list_to_nparray(json_list_input)
        assert converted_input.all() == numpy_array_input.all()

    @pytest.mark.parametrize(
        ("complex_python_input", "complex_json_input"),
        [
            (
                1.23j,
                "1.23j",
            ),
            (
                0j,
                "0j",
            ),
            (
                1.0 + 1.23j,
                "(1+1.23j)",
            ),
            (
                1.1 + 1.23j,
                "(1.1+1.23j)",
            ),
            (
                1 + 2j,
                "(1+2j)",
            ),
            (
                -1.1 - 1.23j,
                "(-1.1-1.23j)",
            ),
            (
                1.0j,
                "1j",
            ),
            (
                1.1 + 0.0j,
                "(1.1+0j)",
            ),
        ],
    )
    def test_complex_python_to_json(self, complex_python_input, complex_json_input):
        """Test conversion from Python complex no. (complex) to JSON compatible string.

        Args:
            complex_python_input (complex): Complex number input.
            complex_json_input (str): Expected JSON compatible output.
        """
        converted_input = conv.complex_python_to_json(complex_python_input)
        assert converted_input == complex_json_input

    @pytest.mark.parametrize(
        ("complex_python_input", "complex_json_input"),
        [
            (
                1.23j,
                "1.23j",
            ),
            (
                0j,
                "0j",
            ),
            (
                1.0 + 1.23j,
                "(1+1.23j)",
            ),
            (
                1.1 + 1.23j,
                "(1.1+1.23j)",
            ),
            (
                1 + 2j,
                "(1+2j)",
            ),
            (
                -1.1 - 1.23j,
                "(-1.1-1.23j)",
            ),
            (
                1.0j,
                "1j",
            ),
            (
                1.1 + 0.0j,
                "(1.1+0j)",
            ),
            (
                1.1 + 1.23j,
                "1.1+1.23j",
            ),
            (
                1.23,
                "1.23",
            ),
            (
                -1j,
                "-j",
            ),
        ],
    )
    def test_complex_json_to_python(self, complex_python_input, complex_json_input):
        """Test conversion from JSON to Python.
        
        Convert JSON compatible string decribing complex no. to Python complex no. (complex).

        Args:
            complex_python_input (complex): Expected complex number output.
            complex_json_input (str):  JSON compatible input.
        """
        converted_input = conv.complex_json_to_python(complex_json_input)
        assert converted_input == complex_python_input

    @pytest.mark.parametrize(
        ("complex_array_input", "complex_json_input"),
        [
            (
                array([1.0 + 1.23j]),
                "[(1+1.23j)]",
            ),
            (
                array([[1.1 + 2.2j, -1.1 + 2.2j], [1.1 - 2.2j, -1.1 - 2.2j]]),
                "[[(1.1+2.2j), (-1.1+2.2j)], [(1.1-2.2j), (-1.1-2.2j)]]",
            ),
        ],
    )
    def test_complex_nparray_to_json_list(
        self, complex_array_input, complex_json_input
    ):
        """Test conversion of NumPy array of type complex to JSON compatible list.

        Args:
            complex_array_input (nparray, dtype=complex): Complex array input.
            complex_json_input (str): Expected JSON compatible list output.
        """
        converted_input = conv.complex_nparray_to_json_list(complex_array_input)
        assert converted_input == complex_json_input



"""Unit tests for JSONDateTimeConverter class."""

from datetime import UTC, datetime, timedelta

import pytest
from qciconnect_common.qciconnect_conversion.json_conversion import JSONDateTimeConverter

conv = JSONDateTimeConverter()


class TestJSONDateTimeConverter:
    """Tests for JSONDateTimeConverter class."""

    @pytest.mark.parametrize(
        ("datetime_input", "datetime_json_input"),
        [
            (
                datetime(2020, 2, 2),
                "2020-02-02T00:00:00.000000",
            ),
            (
                datetime(2020, 2, 2, hour=1, minute=2, second=3),
                "2020-02-02T01:02:03.000000",
            ),
            (
                datetime(2020, 2, 29, hour=1, minute=2, second=3, microsecond=4),
                "2020-02-29T01:02:03.000004",
            ),
            (
                datetime(2020, 2, 29, hour=1, minute=2, second=3, microsecond=4, tzinfo=UTC),
                "2020-02-29T01:02:03.000004",
            ),
        ],
    )
    def test_datetime_to_json_datetime(self, datetime_input: datetime, datetime_json_input: str):
        """Test conversion of Python datetime object to JSON compatible string.

        Args:
            datetime_input (datetime): Python datetime input.
            datetime_json_input (str): Expected JSON compatible output.
        """
        converted_input = conv.datetime_to_json_datetime(datetime_input)
        assert converted_input == datetime_json_input

    @pytest.mark.parametrize(
        ("timedelta_input", "timedelta_json_input"),
        [
            (
                timedelta(microseconds=1),
                "1e-06",
            ),
            (
                timedelta(milliseconds=1),
                "0.001",
            ),
            (
                timedelta(seconds=1),
                "1.0",
            ),
            (
                timedelta(minutes=1),
                "60.0",
            ),
            (
                timedelta(hours=1),
                "3600.0",
            ),
            (
                timedelta(days=1),
                "86400.0",
            ),
            (
                timedelta(weeks=1),
                "604800.0",
            ),
        ],
    )
    def test_timedelta_to_json(self, timedelta_input: timedelta, timedelta_json_input: str):
        """Test conversion of Python datetime.timedelta to JSON compatible value in seconds.

        Args:
            timedelta_input (_type_): Python datetime.timedelta input.
            timedelta_json_input (str): JSON compatible float (as string).
        """
        converted_input = conv.timedelta_to_json(timedelta_input)
        assert converted_input == timedelta_json_input
